# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is configman
#
# The Initial Developer of the Original Code is
# Mozilla Foundation
# Portions created by the Initial Developer are Copyright (C) 2011
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#    K Lars Lohn, lars@mozilla.com
#    Peter Bengtsson, peterbe@mozilla.com
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****

import sys
import os
import os.path
import unittest
from contextlib import contextmanager
import io
from cStringIO import StringIO
import getopt

import mock

import configman.config_manager as config_manager
from configman.option import Option
from configman.dotdict import DotDict, DotDictWithAcquisition
from configman import Namespace, RequiredConfig
from configman.converters import class_converter
import configman.datetime_util as dtu
from configman.config_exceptions import NotAnOptionError
from configman.value_sources.source_exceptions import (
    AllHandlersFailedException,
    UnknownFileExtensionException,
    NoHandlerForType
)


#==============================================================================
class T1(RequiredConfig):
    required_config = Namespace()
    required_config.add_option('a', default=11)


#==============================================================================
class T2(RequiredConfig):
    required_config = Namespace()
    required_config.add_option('b', default=22)


#==============================================================================
class T3(RequiredConfig):
    required_config = Namespace()
    required_config.add_option('c', default=33)
    required_config.namespace('ccc')
    required_config.ccc.add_option('x', default=99)


#==============================================================================
class TestCase(unittest.TestCase):

    #--------------------------------------------------------------------------
    def test_empty_ConfigurationManager_constructor(self):
        # because the default option argument defaults to using sys.argv we
        # have to mock that
        c = config_manager.ConfigurationManager(
            use_admin_controls=False,
            use_auto_help=False,
            argv_source=[]
        )
        self.assertEqual(c.option_definitions, config_manager.Namespace())

    #--------------------------------------------------------------------------
    def test_get_config_1(self):
        n = config_manager.Namespace()
        n.add_option('a', 1, 'the a')
        n.add_option('b', 17)
        c = config_manager.ConfigurationManager(
            [n],
            use_admin_controls=False,
            use_auto_help=False,
            argv_source=[]
        )
        d = c.get_config()
        e = DotDict()
        e.a = 1
        e.b = 17
        self.assertEqual(d, e)

    #--------------------------------------------------------------------------
    def test_get_config_2(self):
        n = config_manager.Namespace()
        n.add_option('a', 1, 'the a')
        n.b = 17
        n.c = c = config_manager.Namespace()
        c.x = 'fred'
        c.y = 3.14159
        c.add_option('z', 99, 'the 99')
        c = config_manager.ConfigurationManager(
            [n],
            use_admin_controls=False,
            use_auto_help=False,
            argv_source=[]
        )
        d = c.get_config()
        e = DotDict()
        e.a = 1
        e.b = 17
        e.c = DotDict()
        e.c.x = 'fred'
        e.c.y = 3.14159
        e.c.z = 99
        self.assertEqual(d, e)

    #--------------------------------------------------------------------------
    def _some_namespaces(self):
        """set up some namespaces"""
        n = config_manager.Namespace(doc='top')
        n.add_option(
            'aaa',
            '2011-05-04T15:10:00',
            'the a',
            short_form='a',
            from_string_converter=dtu.datetime_from_ISO_string
        )
        n.c = config_manager.Namespace(doc='c space')
        n.c.add_option('fred', 'stupid', 'husband from Flintstones')
        n.c.add_option('wilma', 'waspish', 'wife from Flintstones')
        n.d = config_manager.Namespace(doc='d space')
        n.d.add_option('fred', 'crabby', 'male neighbor from I Love Lucy')
        n.d.add_option('ethel', 'silly', 'female neighbor from I Love Lucy')
        n.x = config_manager.Namespace(doc='x space')
        n.x.add_option('size', 100, 'how big in tons', short_form='s')
        n.x.add_option('password', 'secret', 'the password')
        return n

    #--------------------------------------------------------------------------
    def test_overlay_config_4(self):
        """test overlay dict w/flat source dict"""
        n = config_manager.Namespace()
        n.add_option('a', doc='the a', default=1)
        n.b = 17
        n.c = config_manager.Namespace()
        n.c.add_option('extra', doc='the x', default=3.14159)
        g = {'a': 2, 'c.extra': 2.89}
        c = config_manager.ConfigurationManager(
            [n],
            [g],
            use_admin_controls=True,
            use_auto_help=False,
            argv_source=[]
        )
        self.assertTrue(isinstance(c.option_definitions.b,
                                   config_manager.Option))
        self.assertEqual(c.option_definitions.a.value, 2)
        self.assertEqual(c.option_definitions.b.value, 17)
        self.assertEqual(c.option_definitions.b.default, 17)
        self.assertEqual(c.option_definitions.b.name, 'b')
        self.assertEqual(c.option_definitions.c.extra.name, 'extra')
        self.assertEqual(c.option_definitions.c.extra.doc, 'the x')
        self.assertEqual(c.option_definitions.c.extra.default, 2.89)
        self.assertEqual(c.option_definitions.c.extra.value, 2.89)

    #--------------------------------------------------------------------------
    def test_overlay_config_4a(self):
        """test overlay dict w/deep source dict"""
        n = config_manager.Namespace()
        n.add_option('a', 1, doc='the a')
        n.b = 17
        n.c = config_manager.Namespace()
        n.c.add_option('extra', doc='the x', default=3.14159)
        g = {'a': 2, 'c': {'extra': 2.89}}
        c = config_manager.ConfigurationManager(
            [n],
            [g],
            use_admin_controls=True,
            use_auto_help=False,
            argv_source=[]
        )
        self.assertTrue(isinstance(c.option_definitions.b,
                                   config_manager.Option))
        self.assertEqual(c.option_definitions.a.value, 2)
        self.assertEqual(c.option_definitions.b.value, 17)
        self.assertEqual(c.option_definitions.b.default, 17)
        self.assertEqual(c.option_definitions.b.name, 'b')
        self.assertEqual(c.option_definitions.c.extra.name, 'extra')
        self.assertEqual(c.option_definitions.c.extra.doc, 'the x')
        self.assertEqual(c.option_definitions.c.extra.default, 2.89)
        self.assertEqual(c.option_definitions.c.extra.value, 2.89)

    #--------------------------------------------------------------------------
    def test_overlay_config_5(self):
        """test namespace definition w/getopt"""
        n = config_manager.Namespace()
        n.add_option('a', doc='the a', default=1)
        n.b = 17
        n.add_option('c', doc='the c', default=False)
        c = config_manager.ConfigurationManager(
            [n],
            [getopt],
            use_admin_controls=True,
            use_auto_help=False,
            argv_source=['--a', '2', '--c']
        )
        self.assertTrue(isinstance(c.option_definitions.b,
                                   config_manager.Option))
        self.assertEqual(c.option_definitions.a.value, 2)
        self.assertEqual(c.option_definitions.b.value, 17)
        self.assertEqual(c.option_definitions.b.default, 17)
        self.assertEqual(c.option_definitions.b.name, 'b')
        self.assertEqual(c.option_definitions.c.name, 'c')
        self.assertEqual(c.option_definitions.c.value, True)

    #--------------------------------------------------------------------------
    def test_overlay_config_6(self):
        """test namespace definition w/getopt"""
        n = config_manager.Namespace()
        n.add_option('a', doc='the a', default=1)
        n.b = 17
        n.c = config_manager.Namespace()
        n.c.add_option('extra', short_form='e', doc='the x', default=3.14159)
        c = config_manager.ConfigurationManager(
            [n],
            [getopt],
            use_admin_controls=True,
            use_auto_help=False,
            argv_source=['--a', '2', '--c.extra', '11.0']
        )
        self.assertEqual(type(c.option_definitions.b), config_manager.Option)
        self.assertEqual(c.option_definitions.a.value, 2)
        self.assertEqual(c.option_definitions.b.value, 17)
        self.assertEqual(c.option_definitions.b.default, 17)
        self.assertEqual(c.option_definitions.b.name, 'b')
        self.assertEqual(c.option_definitions.c.extra.name, 'extra')
        self.assertEqual(c.option_definitions.c.extra.doc, 'the x')
        self.assertEqual(c.option_definitions.c.extra.default, '11.0')
        self.assertEqual(c.option_definitions.c.extra.value, 11.0)

    #--------------------------------------------------------------------------
    def test_overlay_config_6a(self):
        """test namespace w/getopt w/short form"""
        n = config_manager.Namespace()
        n.add_option('a', doc='the a', default=1)
        n.b = 17
        n.c = config_manager.Namespace()
        n.c.add_option('extra', 3.14159, 'the x', short_form='e')
        c = config_manager.ConfigurationManager(
            [n],
            [getopt],
            use_admin_controls=True,
            use_auto_help=False,
            argv_source=['--a', '2', '-e', '11.0']
        )
        self.assertEqual(type(c.option_definitions.b), config_manager.Option)
        self.assertEqual(c.option_definitions.a.value, 2)
        self.assertEqual(c.option_definitions.b.value, 17)
        self.assertEqual(c.option_definitions.b.default, 17)
        self.assertEqual(c.option_definitions.b.name, 'b')
        self.assertEqual(c.option_definitions.c.extra.name, 'extra')
        self.assertEqual(c.option_definitions.c.extra.doc, 'the x')
        self.assertEqual(c.option_definitions.c.extra.default, '11.0')
        self.assertEqual(c.option_definitions.c.extra.value, 11.0)

    #--------------------------------------------------------------------------
    def test_overlay_config_7(self):
        """test namespace definition flat file"""
        n = config_manager.Namespace()
        n.add_option('a', doc='the a', default=1)
        n.b = 17
        n.c = config_manager.Namespace()
        n.c.add_option('extra', 3.14159, 'the x')
        n.c.add_option('string', 'fred', doc='str')

        @contextmanager
        def dummy_open():
            yield [
                '# comment line to be ignored\n',
                '\n',  # blank line to be ignored
                'a=22\n',
                'b = 33\n',
                'c.extra = 2.0\n',
                'c.string =   wilma\n'
            ]
        c = config_manager.ConfigurationManager(
            [n],
            [dummy_open],
            use_admin_controls=True,
            use_auto_help=False
        )
        self.assertEqual(type(c.option_definitions.b), config_manager.Option)
        self.assertEqual(c.option_definitions.a.value, 22)
        self.assertEqual(c.option_definitions.b.value, 33)
        self.assertEqual(c.option_definitions.b.default, '33')
        self.assertEqual(c.option_definitions.b.name, 'b')
        self.assertEqual(c.option_definitions.c.extra.name, 'extra')
        self.assertEqual(c.option_definitions.c.extra.doc, 'the x')
        self.assertEqual(c.option_definitions.c.extra.default, '2.0')
        self.assertEqual(c.option_definitions.c.extra.value, 2.0)
        self.assertEqual(c.option_definitions.c.string.name, 'string')
        self.assertEqual(c.option_definitions.c.string.doc, 'str')
        self.assertEqual(c.option_definitions.c.string.default, 'wilma')
        self.assertEqual(c.option_definitions.c.string.value, 'wilma')

    #--------------------------------------------------------------------------
    def test_overlay_config_8(self):
        """test namespace definition ini file"""
        n = config_manager.Namespace()
        n.other = config_manager.Namespace()
        n.other.add_option('t', 'tee', 'the t')
        n.d = config_manager.Namespace()
        n.d.add_option('a', 1, doc='the a')
        n.d.b = 17
        n.c = config_manager.Namespace()
        n.c.add_option('extra', 3.14159, 'the x')
        n.c.add_option('string', 'fred', doc='str')
        ini_data = """
other.t=tea
# blank line to be ignored
d.a=22
d.b=33
c.extra = 2.0
c.string = wilma
"""

        def strio():
            return io.BytesIO(ini_data)
        c = config_manager.ConfigurationManager(
            [n], [strio],
            use_admin_controls=True,
            use_auto_help=False
        )
        self.assertEqual(c.option_definitions.other.t.name, 't')
        self.assertEqual(c.option_definitions.other.t.value, 'tea')
        self.assertEqual(type(c.option_definitions.d.b), config_manager.Option)
        self.assertEqual(c.option_definitions.d.a.value, 22)
        self.assertEqual(c.option_definitions.d.b.value, 33)
        self.assertEqual(c.option_definitions.d.b.default, '33')
        self.assertEqual(c.option_definitions.d.b.name, 'b')
        self.assertEqual(c.option_definitions.c.extra.name, 'extra')
        self.assertEqual(c.option_definitions.c.extra.doc, 'the x')
        self.assertEqual(c.option_definitions.c.extra.default, '2.0')
        self.assertEqual(c.option_definitions.c.extra.value, 2.0)
        self.assertEqual(c.option_definitions.c.string.name, 'string')
        self.assertEqual(c.option_definitions.c.string.doc, 'str')
        self.assertEqual(c.option_definitions.c.string.default, 'wilma')
        self.assertEqual(c.option_definitions.c.string.value, 'wilma')

    #--------------------------------------------------------------------------
    def test_overlay_config_9(self):
        """test namespace definition ini file"""
        n = config_manager.Namespace()
        n.other = config_manager.Namespace()
        n.other.add_option('t', 'tee', 'the t')
        n.d = config_manager.Namespace()
        n.d.add_option('a', 1, doc='the a')
        n.d.b = 17
        n.c = config_manager.Namespace()
        n.c.add_option('extra', 3.14159, 'the x')
        n.c.add_option('string', 'fred', 'str')
        ini_data = """
other.t=tea
# blank line to be ignored
d.a=22
c.extra = 2.0
c.string =   from ini
"""

        def strio():
            return io.BytesIO(ini_data)
        e = DotDict()
        e.fred = DotDict()  # should be ignored
        e.fred.t = 'T'  # should be ignored
        e.d = DotDict()
        e.d.a = 16
        e.c = DotDict()
        e.c.extra = 18.6
        e.c.string = 'from environment'

        saved_environ = os.environ
        os.environ = e
        try:
            c = config_manager.ConfigurationManager(
                [n],
                [e, strio, getopt],
                use_admin_controls=True,
                use_auto_help=False,
                argv_source=[
                    '--other.t',
                    'TTT',
                    '--c.extra',
                    '11.0'
                ]
            )
        finally:
            os.environ = saved_environ
        #fm.os = saved_os
        self.assertEqual(c.option_definitions.other.t.name, 't')
        self.assertEqual(c.option_definitions.other.t.value, 'TTT')
        self.assertEqual(type(c.option_definitions.d.b), config_manager.Option)
        self.assertEqual(c.option_definitions.d.a.value, 22)
        self.assertEqual(c.option_definitions.d.b.value, 17)
        self.assertEqual(c.option_definitions.d.b.default, 17)
        self.assertEqual(c.option_definitions.d.b.name, 'b')
        self.assertEqual(c.option_definitions.c.extra.name, 'extra')
        self.assertEqual(c.option_definitions.c.extra.doc, 'the x')
        self.assertEqual(c.option_definitions.c.extra.default, '11.0')
        self.assertEqual(c.option_definitions.c.extra.value, 11.0)
        self.assertEqual(c.option_definitions.c.string.name, 'string')
        self.assertEqual(c.option_definitions.c.string.doc, 'str')
        self.assertEqual(c.option_definitions.c.string.default, 'from ini')
        self.assertEqual(c.option_definitions.c.string.value, 'from ini')

    #--------------------------------------------------------------------------
    def test_overlay_config_10(self):
        """test namespace definition ini file"""
        n = config_manager.Namespace()
        n.other = config_manager.Namespace()
        n.other.add_option('t', 'tee', 'the t')
        n.d = config_manager.Namespace()
        n.d.add_option('a', 1, 'the a')
        n.d.b = 17
        n.c = config_manager.Namespace()
        n.c.add_option('extra', 3.14159, 'the x')
        n.c.add_option('string', 'fred', doc='str')
        ini_data = """
other.t=tea
# blank line to be ignored
d.a=22
c.extra = 2.0
c.string =   from ini
"""

        def strio():
            return io.BytesIO(ini_data)
        e = DotDict()
        e.other = DotDict()
        e.other.t = 'T'
        e.d = DotDict()
        e.d.a = 16
        e.c = DotDict()
        e.c.extra = 18.6
        e.c.string = 'from environment'

        c = config_manager.ConfigurationManager(
            [n],
            [e, strio, getopt],
            use_admin_controls=True,
            argv_source=['--c.extra', '11.0'],
            use_auto_help=False
        )
        self.assertEqual(c.option_definitions.other.t.name, 't')
        self.assertEqual(c.option_definitions.other.t.value, 'tea')
        self.assertEqual(type(c.option_definitions.d.b), config_manager.Option)
        self.assertEqual(c.option_definitions.d.a.value, 22)
        self.assertEqual(c.option_definitions.d.b.value, 17)
        self.assertEqual(c.option_definitions.d.b.default, 17)
        self.assertEqual(c.option_definitions.d.b.name, 'b')
        self.assertEqual(c.option_definitions.c.extra.name, 'extra')
        self.assertEqual(c.option_definitions.c.extra.doc, 'the x')
        self.assertEqual(c.option_definitions.c.extra.default, '11.0')
        self.assertEqual(c.option_definitions.c.extra.value, 11.0)
        self.assertEqual(c.option_definitions.c.string.name, 'string')
        self.assertEqual(c.option_definitions.c.string.doc, 'str')
        self.assertEqual(c.option_definitions.c.string.default, 'from ini')
        self.assertEqual(c.option_definitions.c.string.value, 'from ini')

    #--------------------------------------------------------------------------
    def test_overlay_config_11(self):
        """test overlay dict w/deep source dict and reference value links"""
        n = config_manager.Namespace()
        n.add_option('a', 1, doc='the a', reference_value_from='xxx.yyy')
        n.b = 17
        n.c = config_manager.Namespace()
        n.c.add_option(
            'extra', doc='the x',
            default=3.14159,
            reference_value_from='xxx.yyy'
        )
        n.c.add_option(
            'a',
            doc='the a',
            reference_value_from='xxx.yyy'
        )
        g = {
            'xxx': {
                'yyy': {
                    'a': 2,
                    'extra': 2.89
                }
            }
        }
        c = config_manager.ConfigurationManager(
            [n],
            [g],
            use_admin_controls=True,
            use_auto_help=False,
            argv_source=[]
        )
        self.assertTrue(isinstance(
            c.option_definitions.b,
            config_manager.Option
        ))
        self.assertEqual(c.option_definitions.a.value, 2)
        self.assertEqual(c.option_definitions.b.value, 17)
        self.assertEqual(c.option_definitions.b.default, 17)
        self.assertEqual(c.option_definitions.b.name, 'b')
        self.assertEqual(c.option_definitions.c.extra.name, 'extra')
        self.assertEqual(c.option_definitions.c.extra.doc, 'the x')
        self.assertEqual(c.option_definitions.c.extra.default, 2.89)
        self.assertEqual(c.option_definitions.c.extra.value, 2.89)
        self.assertEqual(c.option_definitions.c.a.value, 2)

    #--------------------------------------------------------------------------
    def test_mapping_types_1(self):
        n = config_manager.Namespace()
        n.add_option(
            name='a',
            default=1,
            doc='the a'
        )
        n.b = 17
        n.c = config_manager.Namespace()
        n.c.x = 'fred'
        n.c.y = 3.14159
        n.c.add_option(
            name='z',
            default=99,
            doc='the 99'
        )
        o = {"a": 2, "c.z": 22, "c.x": 'noob', "c.y": "2.89"}
        c = config_manager.ConfigurationManager(
            [n],
            [o],
            use_admin_controls=False,
            use_auto_help=False,
            argv_source=[]
        )
        e = DotDict()
        e.a = 2
        e.b = 17
        e.c = DotDict()
        e.c.x = 'noob'
        e.c.y = 2.89
        e.c.z = 22
        d = c._generate_config(dict)
        self.assertTrue(isinstance(d, dict))
        self.assertTrue(isinstance(d['c'], dict))
        self.assertEqual(d, e)
        d = c._generate_config(DotDict)
        self.assertTrue(isinstance(d, DotDict))
        self.assertTrue(isinstance(d.c, DotDict))
        self.assertEqual(d, e)
        d = c._generate_config(DotDictWithAcquisition)
        self.assertTrue(isinstance(d, DotDictWithAcquisition))
        self.assertTrue(isinstance(d.c, DotDictWithAcquisition))
        self.assertEqual(d, e)
        self.assertEqual(d.a, 2)
        self.assertEqual(d.c.a, 2)
        self.assertEqual(d.c.b, 17)

    #--------------------------------------------------------------------------
    def test_get_option_names(self):
        n = config_manager.Namespace()
        n.add_option('a', 1, 'the a')
        n.b = 17
        n.c = config_manager.Namespace()
        n.c.add_option('fred')
        n.c.add_option('wilma')
        n.d = config_manager.Namespace()
        n.d.add_option('fred')
        n.d.add_option('wilma')
        n.d.x = config_manager.Namespace()
        n.d.x.add_option('size')
        c = config_manager.ConfigurationManager(
            [n],
            use_admin_controls=False,
            use_auto_help=False,
            argv_source=[]
        )
        names = c.get_option_names()
        names.sort()
        e = ['a', 'b', 'c.fred', 'c.wilma', 'd.fred', 'd.wilma', 'd.x.size']
        e.sort()
        self.assertEqual(names, e)
        names = sorted([x for x in c.option_definitions.keys_breadth_first()])
        self.assertEqual(names, e)

    #--------------------------------------------------------------------------
    def test_get_option(self):
        n = config_manager.Namespace()
        n.add_option('a', 1, 'the a')
        n.b = 17
        n.c = config_manager.Namespace()
        n.c.add_option('fred')
        n.c.add_option('wilma')
        n.d = config_manager.Namespace()
        n.d.add_option('fred')
        n.d.add_option('wilma')
        n.d.x = config_manager.Namespace()
        n.d.x.add_option('size')
        c = config_manager.ConfigurationManager(
            [n],
            use_admin_controls=True,
            use_auto_help=False,
            argv_source=[]
        )
        self.assertEqual(c._get_option('a'), n.a)
        self.assertEqual(c._get_option('b').name, 'b')
        self.assertEqual(c._get_option('c.fred'), n.c.fred)
        self.assertEqual(c._get_option('c.wilma'), n.c.wilma)
        self.assertEqual(c._get_option('d.fred'), n.d.fred)
        self.assertEqual(c._get_option('d.wilma'), n.d.wilma)
        self.assertEqual(c._get_option('d.wilma'), n.d.wilma)
        self.assertEqual(c._get_option('d.x.size'), n.d.x.size)

    #--------------------------------------------------------------------------
    def test_output_summary(self):
        """test_output_summary: the output from help"""
        n = config_manager.Namespace()
        n.add_option('aaa', False, 'the a', short_form='a')
        n.add_option('bee', True)
        n.b = 17
        n.c = config_manager.Namespace()
        n.c.add_option('fred', doc='husband from Flintstones')
        n.d = config_manager.Namespace()
        n.d.add_option('fred', doc='male neighbor from I Love Lucy')
        n.d.x = config_manager.Namespace()
        n.d.x.add_option('size', 100, 'how big in tons', short_form='s')
        n.d.x.add_option('password', 'secrets', 'the password')
        c = config_manager.ConfigurationManager(
            [n],
            use_admin_controls=True,
            use_auto_help=False,
            argv_source=[],
        )
        s = StringIO()
        c.output_summary(output_stream=s)
        r = s.getvalue()
        self.assertTrue('OPTIONS:\n' in r)

        options = r.split('OPTIONS:\n')[1]
        s.close()

        padding = '\n' + ' ' * 4
        expect = [
            ('-a, --aaa', 'the a%s(default: False)' % padding),
            ('--b', '(default: 17)'),
            ('--bee', '(default: True)'),
            ('--c.fred', 'husband from Flintstones'),
            ('--d.fred', 'male neighbor from I Love Lucy'),
            ('--d.x.password', 'the password%s(default: *********)' % padding),
            ('-s, --d.x.size', 'how big in tons%s(default: 100)' % padding),
        ]
        point = -1  # used to assert the sort order
        for i, (start, end) in enumerate(expect):
            self.assertTrue(
                point < options.find(start) < options.find(end),
                expect[i]
            )
            point = options.find(end)

    #--------------------------------------------------------------------------
    def test_output_summary_header(self):
        """a config with an app_name, app_version and app_description is
        printed on the output summary.
        """
        n = config_manager.Namespace()
        n.add_option('aaa', False, 'the a', short_form='a')
        c = config_manager.ConfigurationManager(
            n,
            use_admin_controls=True,
            use_auto_help=False,
            argv_source=[],
        )

        def get_output(conf):
            s = StringIO()
            conf.output_summary(output_stream=s)
            return s.getvalue()

        output = get_output(c)
        assert 'OPTIONS:' in output
        self.assertTrue('Application:' not in output)

        c.app_name = 'foobar'
        output = get_output(c)
        assert 'OPTIONS:' in output
        self.assertTrue('Application: foobar' in output)

        c.app_version = '1.0'
        output = get_output(c)
        assert 'OPTIONS:' in output
        self.assertTrue('Application: foobar 1.0' in output)

        c.app_description = "This ain't your mama's app"
        output = get_output(c)
        assert 'OPTIONS:' in output
        self.assertTrue('Application: foobar 1.0\n' in output)
        self.assertTrue("This ain't your mama's app\n\n" in output)

    #--------------------------------------------------------------------------
    def test_eval_as_converter(self):
        """does eval work as a to string converter on an Option object?"""
        n = config_manager.Namespace()
        n.add_option('aaa', doc='the a', default='', short_form='a')
        self.assertEqual(n.aaa.value, '')

    #--------------------------------------------------------------------------
    def test_RequiredConfig_get_required_config(self):

        class Foo:
            required_config = {'foo': True}

        class Bar:
            required_config = {'bar': False}

        class Poo:
            pass

        class Combined(config_manager.RequiredConfig, Foo, Poo, Bar):
            pass

        result = Combined.get_required_config()
        self.assertEqual(result.foo.value, True)
        self.assertEqual(result.bar.value, False)

        c = Combined()
        c.config_assert({'foo': True, 'bar': False})

        self.assertRaises(AssertionError, c.config_assert, ({},))

    #--------------------------------------------------------------------------
    def test_app_name_from_app_obj(self):
        class MyApp(config_manager.RequiredConfig):
            app_name = 'fred'
            app_version = '1.0'
            app_description = "my app"

            def __init__(inner_self, config):
                inner_self.config = config

        n = config_manager.Namespace()
        n.admin = config_manager.Namespace()
        n.add_option(
            'application',
            MyApp,
            'the app object class'
        )
        c = config_manager.ConfigurationManager(
            [n],
            use_admin_controls=True,
            use_auto_help=False,
            argv_source=[]
        )
        self.assertEqual(c.app_name, MyApp.app_name)
        self.assertEqual(c.app_version, MyApp.app_version)
        self.assertEqual(c.app_description, MyApp.app_description)

    #--------------------------------------------------------------------------
    def test_help_out(self):
        class MyApp(config_manager.RequiredConfig):
            app_name = 'fred'
            app_version = '1.0'
            app_description = "my app"
            required_config = config_manager.Namespace()
            required_config.add_option(
                'password',
                'fred',
                'the password',
                short_form='p'
            )
            required_config.add_option(
                'password_wo_default',
                doc='This one has no default'
            )
            required_config.add_option(
                'password_wo_docstr',
                default='Something'
            )

            def __init__(inner_self, config):
                inner_self.config = config

        n = config_manager.Namespace()
        n.admin = config_manager.Namespace()
        n.add_option(
            'application',
            MyApp,
            'the app object class',
            short_form='a'
        )

        class MyConfigManager(config_manager.ConfigurationManager):
            def output_summary(inner_self):
                output_stream = StringIO()
                r = super(MyConfigManager, inner_self).output_summary(
                    output_stream=output_stream,
                    block_password=False
                )
                r = output_stream.getvalue()
                output_stream.close()
                self.assertTrue('Application: fred 1.0' in r)
                self.assertTrue('my app\n\n' in r)
                self.assertTrue('OPTIONS:\n' in r)
                self.assertTrue('  --help' in r and 'print this' in r)
                self.assertTrue('print this (default: True)' not in r)
                self.assertTrue('  --password' in r)
                self.assertTrue('(default: *********)' in r)
                self.assertTrue('  --application' not in r)

        def my_exit():
            pass

        old_sys_exit = sys.exit
        sys.exit = my_exit
        try:
            MyConfigManager(
                n,
                [getopt],
                use_admin_controls=True,
                use_auto_help=True,
                argv_source=['--password=wilma', '--help']
            )
        finally:
            sys.exit = old_sys_exit

    #--------------------------------------------------------------------------
    def test_write_gets_called(self):
        class MyApp(config_manager.RequiredConfig):
            app_name = 'fred'
            app_version = '1.0'
            app_description = "my app"
            required_config = config_manager.Namespace()
            required_config.add_option('password', 'fred', 'the password')

            def __init__(inner_self, config):
                inner_self.config = config

        n = config_manager.Namespace()
        n.admin = config_manager.Namespace()
        n.add_option(
            'application',
            MyApp,
            'the app object class'
        )

        class MyConfigManager(config_manager.ConfigurationManager):
            def __init__(inner_self, *args, **kwargs):
                inner_self.write_called = False
                super(MyConfigManager, inner_self).__init__(*args, **kwargs)

            def dump_conf(inner_self):
                inner_self.dump_conf_called = True

        def my_exit():
            pass
        old_sys_exit = sys.exit
        sys.exit = my_exit
        try:
            c = MyConfigManager(
                n,
                [getopt],
                use_admin_controls=True,
                use_auto_help=True,
                argv_source=[
                    '--password=wilma',
                    '--admin.dump_conf=x.ini'
                ]
            )
            self.assertEqual(c.dump_conf_called, True)
        finally:
            sys.exit = old_sys_exit

    #--------------------------------------------------------------------------
    def test_get_options(self):
        class MyApp(config_manager.RequiredConfig):
            app_name = 'fred'
            app_version = '1.0'
            app_description = "my app"
            required_config = config_manager.Namespace()
            required_config.add_option('password', 'fred', 'the password')
            required_config.sub = config_manager.Namespace()
            required_config.sub.add_option(
                'name',
                'ethel',
                'the name'
            )

            def __init__(inner_self, config):
                inner_self.config = config

        n = config_manager.Namespace()
        n.admin = config_manager.Namespace()
        n.add_option(
            'application',
            MyApp,
            'the app object class'
        )

        c = config_manager.ConfigurationManager(
            n,
            use_admin_controls=True,
            use_auto_help=False,
            argv_source=[]
        )
        r = sorted(c._get_options())
        e = sorted([
            ('admin.print_conf', 'print_conf', None),
            ('admin.dump_conf', 'dump_conf', ''),
            ('admin.conf', 'conf', None),
            ('admin.strict', 'strict', False),
            ('application', 'application', MyApp),
            ('password', 'password', 'fred'),
            ('sub.name', 'name', 'ethel')
        ])
        for expected, result in zip(e, r):
            expected_key, expected_name, expected_default = expected
            result_key, result_option = result
            self.assertEqual(expected_key, result_key)
            self.assertEqual(expected_name, result_option.name)
            self.assertEqual(expected_default, result_option.default)

    #--------------------------------------------------------------------------
    def test_log_config(self):
        class MyApp(config_manager.RequiredConfig):
            app_name = 'fred'
            app_version = '1.0'
            app_description = "my app"
            required_config = config_manager.Namespace()
            required_config.add_option('password', 'fred', 'the password')
            required_config.sub = config_manager.Namespace()
            required_config.sub.add_option(
                'name',
                'ethel',
                'the name'
            )

            def __init__(inner_self, config):
                inner_self.config = config

        n = config_manager.Namespace()
        n.add_option(
            'application',
            MyApp,
            'the app object class'
        )

        c = config_manager.ConfigurationManager(
            n,
            [getopt],
            use_admin_controls=True,
            use_auto_help=False,
            argv_source=['--sub.name=wilma']
        )

        class FakeLogger(object):
            def __init__(self):
                self.log = []

            def info(self, *args):
                self.log.append(args[0] % args[1:])

        fl = FakeLogger()
        c.log_config(fl)
        e = [
            "app_name: fred",
            "app_version: 1.0",
            "current configuration:",
            "application: <class 'configman.tests.test_config_manager.MyApp'>",
            "password: *********",
            "sub.name: wilma"
        ]
        for expected, received in zip(e, fl.log):
            self.assertEqual(expected, received)

    #--------------------------------------------------------------------------
    def test_extra_commandline_parameters(self):
        class MyApp(config_manager.RequiredConfig):
            app_name = 'fred'
            app_version = '1.0'
            app_description = "my app"
            required_config = config_manager.Namespace()
            required_config.add_option('password', 'fred', 'the password')
            required_config.sub = config_manager.Namespace()
            required_config.sub.add_option(
                'name',
                'ethel',
                'the name'
            )

            def __init__(inner_self, config):
                inner_self.config = config

        n = config_manager.Namespace()
        n.admin = config_manager.Namespace()
        n.add_option(
            'application',
            MyApp,
            'the app object class'
        )

        c = config_manager.ConfigurationManager(
            n,
            [getopt],
            use_admin_controls=True,
            use_auto_help=False,
            argv_source=[
                '--sub.name=wilma',
                'argument 1',
                'argument 2',
                'argument 3'
            ]
        )
        expected = ['argument 1',
                    'argument 2',
                    'argument 3']
        self.assertEqual(c.args, expected)

    #--------------------------------------------------------------------------
    def test_print_conf_called(self):
        class MyApp(config_manager.RequiredConfig):
            app_name = 'fred'
            app_version = '1.0'
            app_description = "my app"
            required_config = config_manager.Namespace()
            required_config.add_option('password', 'fred', 'the password')
            required_config.sub = config_manager.Namespace()
            required_config.sub.add_option(
                'name',
                'ethel',
                'the name'
            )

            def __init__(inner_self, config):
                inner_self.config = config

        n = config_manager.Namespace()
        n.admin = config_manager.Namespace()
        n.add_option(
            'application',
            MyApp,
            'the app object class'
        )

        class MyConfigManager(config_manager.ConfigurationManager):
            def __init__(inner_self, *args, **kwargs):
                inner_self.write_called = False
                super(MyConfigManager, inner_self).__init__(*args, **kwargs)

            def print_conf(inner_self):
                inner_self.print_conf_called = True

        c = MyConfigManager(n,
                            [getopt],
                            use_admin_controls=True,
                            use_auto_help=False,
                            quit_after_admin=False,
                            argv_source=['--admin.print_conf=ini',
                                         'argument 1',
                                         'argument 2',
                                         'argument 3'])
        self.assertEqual(c.print_conf_called, True)

    #--------------------------------------------------------------------------
    def test_non_compliant_app_object(self):
        # the MyApp class doesn't define required config
        class MyApp():
            app_name = 'fred'
            app_version = '1.0'
            app_description = "my app"

            def __init__(inner_self, config):
                inner_self.config = config

        n = config_manager.Namespace()
        n.admin = config_manager.Namespace()
        n.add_option(
            'application',
            MyApp,
            'the app object class'
        )

        c = config_manager.ConfigurationManager(
            n,
            [getopt],
            use_admin_controls=True,
            use_auto_help=False,
            argv_source=[
                'argument 1',
                'argument 2',
                'argument 3'
            ]
        )
        conf = c.get_config()
        self.assertEqual(conf.keys(), ['admin', 'application'])

    #--------------------------------------------------------------------------
    def test_print_conf(self):
        n = config_manager.Namespace()

        class MyConfigManager(config_manager.ConfigurationManager):
            def __init__(inner_self, *args, **kwargs):
                inner_self.write_called = False
                super(MyConfigManager, inner_self).__init__(*args, **kwargs)

            def print_conf(self):
                temp_stdout = sys.stdout
                sys.stdout = 17
                try:
                    super(MyConfigManager, self).print_conf()
                finally:
                    sys.stdout = temp_stdout

            def write_conf(inner_self, file_type, opener, skip_keys=None):
                self.assertEqual(file_type, 'ini')
                with opener() as f:
                    self.assertEqual(f, 17)

        MyConfigManager(
            n,
            [getopt],
            use_admin_controls=True,
            use_auto_help=False,
            quit_after_admin=False,
            argv_source=['--admin.print_conf=ini',
                         'argument 1',
                         'argument 2',
                         'argument 3'],
            config_pathname='fred')

    #--------------------------------------------------------------------------
    def test_dump_conf(self):
        n = config_manager.Namespace()

        class MyConfigManager(config_manager.ConfigurationManager):
            def __init__(inner_self, *args, **kwargs):
                inner_self.write_called = False
                super(MyConfigManager, inner_self).__init__(*args, **kwargs)

            def write_conf(inner_self, file_type, opener, skip_keys=None):
                self.assertEqual(file_type, 'ini')
                self.assertEqual(opener.args, ('fred.ini', 'w'))

        MyConfigManager(
            n,
            [getopt],
            use_admin_controls=True,
            use_auto_help=False,
            quit_after_admin=False,
            argv_source=['--admin.dump_conf=fred.ini',
                         'argument 1',
                         'argument 2',
                         'argument 3'],
            config_pathname='fred'
        )

    #--------------------------------------------------------------------------
    def test_dump_conf_bad_extension(self):
        n = config_manager.Namespace()

        self.assertRaises(
            UnknownFileExtensionException,
            config_manager.ConfigurationManager,
            n,
            [getopt],
            use_admin_controls=True,
            use_auto_help=False,
            quit_after_admin=False,
            argv_source=[
                '--admin.dump_conf=/tmp/fred.xxx',
                'argument 1',
                'argument 2',
                'argument 3'
            ],
            config_pathname='fred'
        )
        self.assertFalse(os.path.exists('/tmp/fred.xxx'))

    #--------------------------------------------------------------------------
    def test_print_conf_some_options_excluded(self):
        n = config_manager.Namespace()
        n.add_option('gender',
                     default='Male',
                     doc='What kind of genitalia?')
        n.add_option('salary',
                     default=10000,
                     doc='How much do you earn?',
                     exclude_from_print_conf=True
                     )

        old_stdout = sys.stdout
        temp_output = StringIO()
        sys.stdout = temp_output
        try:
            config_manager.ConfigurationManager(
                n,
                [getopt],
                use_admin_controls=True,
                use_auto_help=False,
                quit_after_admin=False,
                argv_source=['--admin.print_conf=ini'],
                config_pathname='fred'
            )
        finally:
            sys.stdout = old_stdout

        printed = temp_output.getvalue()
        self.assertTrue('gender' in printed)
        self.assertTrue('salary' not in printed)

    #--------------------------------------------------------------------------
    def test_dump_conf_some_options_excluded(self):
        n = config_manager.Namespace()
        n.add_option('gender',
                     default='Male',
                     doc='What kind of genitalia?',
                     exclude_from_print_conf=True)
        n.add_option('salary',
                     default=10000,
                     doc='How much do you earn?',
                     exclude_from_dump_conf=True
                     )

        try:
            config_manager.ConfigurationManager(
                n,
                [getopt],
                use_admin_controls=True,
                use_auto_help=False,
                quit_after_admin=False,
                argv_source=['--admin.dump_conf=foo.conf'],
                config_pathname='fred'
            )

            printed = open('foo.conf').read()
            self.assertTrue('gender' in printed)
            self.assertTrue('salary' not in printed)

        finally:
            if os.path.isfile('foo.conf'):
                os.remove('foo.conf')

    #--------------------------------------------------------------------------
    def test_ConfigurationManager_block_password(self):
        function = config_manager.ConfigurationManager._block_password
        self.assertEqual(function('foo', 'bar', 'peter', block_password=False),
                         ('foo', 'bar', 'peter'))
        self.assertEqual(function('foo', 'bar', 'peter', block_password=True),
                         ('foo', 'bar', 'peter'))
        self.assertEqual(function('foo', 'password', 'peter',
                                  block_password=True),
                         ('foo', 'password', '*********'))
        self.assertEqual(function('foo', 'my_password', 'peter',
                                  block_password=True),
                         ('foo', 'my_password', '*********'))

    #--------------------------------------------------------------------------
    def test_do_aggregations(self):
        def aggregation_test(all_config, local_namespace, args):
            self.assertTrue('password' in all_config)
            self.assertTrue('sub1' in all_config)
            self.assertTrue('name' in all_config.sub1)
            self.assertTrue('name' in local_namespace)
            self.assertTrue('spouse' in local_namespace)
            self.assertEqual(len(args), 2)
            return (
                '%s married %s using password %s but divorced because of %s.'
                % (
                    local_namespace.name,
                    local_namespace.spouse,
                    all_config.password,
                    args[1]
                )
            )

        class MyApp(config_manager.RequiredConfig):
            app_name = 'fred'
            app_version = '1.0'
            app_description = "my app"
            required_config = config_manager.Namespace()
            required_config.add_option('password', '@$*$&26Ht', 'the password')
            required_config.namespace('sub1')
            required_config.sub1.add_option('name', 'ethel', 'the name')
            required_config.sub1.add_option('spouse', 'fred', 'the spouse')
            required_config.sub1.add_aggregation('statement', aggregation_test)

            def __init__(inner_self, config):
                inner_self.config = config

        n = config_manager.Namespace()
        n.admin = config_manager.Namespace()
        n.add_option(
            'application',
            MyApp,
            'the app object class'
        )

        c = config_manager.ConfigurationManager(
            n,
            [getopt],
            use_admin_controls=True,
            use_auto_help=False,
            argv_source=[
                '--sub1.name=wilma',
                'arg1',
                'arg2'
            ]
        )
        config = c.get_config()
        self.assertEqual(config.sub1.statement,
                         'wilma married fred using password @$*$&26Ht '
                         'but divorced because of arg2.')

    #--------------------------------------------------------------------------
    def test_context(self):

        class AggregatedValue(object):

            def __init__(self, value):
                self.value = value

            def close(self):
                self.value = None

        def aggregation_test(all_config, local_namespace, args):
            self.assertTrue('password' in all_config)
            self.assertTrue('sub1' in all_config)
            self.assertTrue('name' in all_config.sub1)
            self.assertTrue('name' in local_namespace)
            self.assertTrue('spouse' in local_namespace)
            self.assertEqual(len(args), 2)
            return AggregatedValue('%s married %s using password %s but '
                                   'divorced because of %s.' %
                                   (local_namespace.name,
                                    local_namespace.spouse,
                                    all_config.password,
                                    args[1]))

        class MyApp(config_manager.RequiredConfig):
            app_name = 'fred'
            app_version = '1.0'
            app_description = "my app"
            required_config = config_manager.Namespace()
            required_config.add_option('password', '@$*$&26Ht', 'the password')
            required_config.namespace('sub1')
            required_config.sub1.add_option('name', 'ethel', 'the name')
            required_config.sub1.add_option('spouse', 'fred', 'the spouse')
            required_config.sub1.add_aggregation('statement', aggregation_test)

            def __init__(inner_self, config):
                inner_self.config = config

        n = config_manager.Namespace()
        n.add_option('application',
                     MyApp,
                     'the app object class')

        c = config_manager.ConfigurationManager(
            n,
            [getopt],
            use_admin_controls=False,
            use_auto_help=False,
            argv_source=[
                '--sub1.name=wilma',
                'arg1',
                'arg2'
            ]
        )

        with c.context() as config:
            statement = config.sub1.statement
            self.assertEqual(statement.value,
                             'wilma married fred using password @$*$&26Ht '
                             'but divorced because of arg2.')
        self.assertTrue(statement.value is None)

    #--------------------------------------------------------------------------
    def test_failing_aggregate_error_bubbling(self):
        """Reproduces and assures this issue
        https://github.com/mozilla/configman/issues/21
        """
        class AggregatedValue(object):

            def __init__(self, value):
                self.value = value

            def close(self):
                self.value = None

        class SomeException(Exception):
            pass

        def aggregation_test(all_config, local_namespace, args):
            # the aggregator might be broken
            raise SomeException('anything')

        class MyApp(config_manager.RequiredConfig):
            app_name = 'fred'
            app_version = '1.0'
            app_description = "my app"
            required_config = config_manager.Namespace()
            required_config.add_aggregation('statement', aggregation_test)

        n = config_manager.Namespace()
        n.admin = config_manager.Namespace()
        n.add_option(
            'application',
            MyApp,
            'the app object class'
        )

        c = config_manager.ConfigurationManager(
            n,
            [getopt],
            use_admin_controls=True,
            use_auto_help=False,
            argv_source=[]
        )

        contextmanager_ = c.context()
        self.assertRaises(SomeException, contextmanager_.__enter__)

    #--------------------------------------------------------------------------
    def test_namespaces_with_conflicting_class_converters(self):
        rc = Namespace()
        rc.namespace('source')
        rc.source.add_option(
            'cls',
            default='configman.tests.test_config_manager.T1',
            from_string_converter=class_converter
        )
        rc.namespace('destination')
        rc.destination.add_option(
            'cls',
            default='configman.tests.test_config_manager.T2',
            from_string_converter=class_converter
        )
        c = config_manager.ConfigurationManager(
            rc,
            [
                {
                    'source': {
                        'cls': 'configman.tests.test_config_manager.T2'
                    },
                    'destination': {
                        'cls': 'configman.tests.test_config_manager.T3'
                    }
                },
                {
                    'source': {
                        'cls': 'configman.tests.test_config_manager.T1'
                    },
                    'destination': {
                        'cls': 'configman.tests.test_config_manager.T2'
                    }
                },
                {
                    'source': {
                        'cls': 'configman.tests.test_config_manager.T3'
                    },
                    'destination': {
                        'cls': 'configman.tests.test_config_manager.T1'
                    }
                },
            ],
            use_admin_controls=True,
            use_auto_help=False,
            argv_source=[]
        )
        conf = c.get_config()
        self.assertEqual(len(conf), 3)
        self.assertEqual(conf.keys(), ['source', 'destination', 'admin'])
        self.assertEqual(len(conf.source), 3)
        self.assertEqual(conf.source.c, 33)
        self.assertEqual(conf.source.cls, T3)
        self.assertEqual(len(conf.source.ccc), 1)
        self.assertEqual(conf.source.ccc.x, 99)
        self.assertEqual(len(conf.destination), 2)
        self.assertEqual(conf.destination.a, 11)
        self.assertEqual(conf.destination.cls, T1)

    #--------------------------------------------------------------------------
    def _common_app_namespace_setup(self):
        class MyApp(config_manager.RequiredConfig):
            app_name = 'fred'
            app_version = '1.0'
            app_description = "my app"
            required_config = config_manager.Namespace()
            required_config.namespace('toplevel')
            required_config.toplevel.add_option('password', 'fred',
                                                'the password')

        n = config_manager.Namespace()
        n.admin = config_manager.Namespace()
        n.add_option(
            'application',
            MyApp,
            'the app object class'
        )
        return n

    #--------------------------------------------------------------------------
    def test_admin_conf_missing_file_ioerror(self):
        """if you specify an `--admin.conf=...` file that doesn't exist it
        should not let you get away with it.
        """
        n = self._common_app_namespace_setup()

        self.assertRaises(
            IOError,
            config_manager.ConfigurationManager,
            (n,),
            argv_source=['--admin.conf=x.ini']
        )

        # but check we can still do it if the file exists
        open('x.ini', 'w').write(
            '[toplevel]\n'
            'password=something\n'
        )
        try:
            c = config_manager.ConfigurationManager(
                (n,),
                argv_source=['--admin.conf=x.ini']
            )
            with c.context() as config:
                self.assertEqual(config.toplevel.password, 'something')
        finally:
            os.remove('x.ini')

    #--------------------------------------------------------------------------
    def test_bad_options(self):
        """tests _check_for_mismatches"""
        rc = Namespace()
        rc.namespace('source')
        rc.source.add_option(
            'cls',
            default='configman.tests.test_config_manager.T1',
            from_string_converter=class_converter
        )
        rc.namespace('destination')
        rc.destination.add_option(
            'cls',
            default='configman.tests.test_config_manager.T2',
            from_string_converter=class_converter
        )
        self.assertRaises(  # 'classy' is not an option
            NotAnOptionError,
            config_manager.ConfigurationManager,
            rc,
            [
                {
                    'admin': {
                        'strict': True
                    }
                },
                {
                    'source': {
                        'cls': 'configman.tests.test_config_manager.T2'
                    },
                    'destination': {
                        'cls': 'configman.tests.test_config_manager.T3'
                    }
                },
                {
                    'source': {
                        'cls': 'configman.tests.test_config_manager.T1'
                    },
                    'destination': {
                        'cls': 'configman.tests.test_config_manager.T2'
                    }
                },
                {
                    'source': {
                        'classy': 'configman.tests.test_config_manager.T3'
                    },
                    'destination': {
                        'cls': 'configman.tests.test_config_manager.T1'
                    }
                },
            ],
        )
        self.assertRaises(  # 'sourness' not a namespace
            NotAnOptionError,
            config_manager.ConfigurationManager,
            rc,
            [
                {
                    'admin': {
                        'strict': True
                    }
                },
                {
                    'source': {
                        'cls': 'configman.tests.test_config_manager.T2'
                    },
                    'destination': {
                        'cls': 'configman.tests.test_config_manager.T3'
                    }
                },
                {
                    'sourness': {
                        'cls': 'configman.tests.test_config_manager.T1'
                    },
                    'destination': {
                        'cls': 'configman.tests.test_config_manager.T2'
                    }
                },
                {
                    'source': {
                        'cls': 'configman.tests.test_config_manager.T3'
                    },
                    'destination': {
                        'cls': 'configman.tests.test_config_manager.T1'
                    }
                },
            ],
        )

        # make sure commandline data sources always behave stictly even when
        # strict is set to False
        import getopt
        self.assertRaises(  # 'alpha' is not an option
            NotAnOptionError,
            config_manager.ConfigurationManager,
            rc,
            [
                {'admin': {'strict': False}},  # not strict, we allow anything
                {
                    'source': {
                        'cls': 'configman.tests.test_config_manager.T2'
                    },
                    'destination': {
                        'cls': 'configman.tests.test_config_manager.T3'
                    }
                },
                {
                    'source': {
                        'cls': 'configman.tests.test_config_manager.T1'
                    },
                    'destination': {
                        'cls': 'configman.tests.test_config_manager.T2'
                    }
                },
                # commandline sources need to ignore when strict=False and
                # raise NotAnOptionError anyway - we don't want to allow
                # arbitrary  switches on the command line.
                getopt,
            ],
            argv_source=['--alpha']
        )

    #--------------------------------------------------------------------------
    def test_acquisition(self):
        """define a common key in two sub-namespaces.  Then offer only a value
        from the base namespace.  Both sub-namespace Options should have the
        end value from the base value namespace."""
        rc = Namespace()
        rc.namespace('source')
        rc.source.add_option('cls',
                             default='configman.tests.test_config_manager.T1',
                             from_string_converter=class_converter)
        rc.namespace('destination')
        rc.destination.add_option(
            'cls',
            default='configman.tests.test_config_manager.T2',
            from_string_converter=class_converter
        )
        cm = config_manager.ConfigurationManager(
            rc,
            [
                {'cls': 'configman.tests.test_config_manager.T2'},
            ],
        )
        c = cm.get_config()
        self.assertEqual(c.source.cls, T2)
        self.assertEqual(c.destination.cls, T2)

    #--------------------------------------------------------------------------
    def test_admin_conf_all_handlers_fail(self):
        """no handler found produces empty message"""
        n = self._common_app_namespace_setup()

        # make a config file that nothing will understand
        open('x.fred', 'w').write(
            '[toplevel]\n'
            'password=something\n'
        )
        try:
            self.assertRaises(
                NoHandlerForType,
                config_manager.ConfigurationManager,
                (n,),
                argv_source=['--admin.conf=x.fred']
            )
        finally:
            os.remove('x.fred')

    #--------------------------------------------------------------------------
    def test_admin_conf_fail_message_gets_through(self):
        """make sure AllHandlersFailedException message gets through"""
        n = self._common_app_namespace_setup()

        # make a config file that fails to load in its proper handler
        open('x.ini', 'w').write(
            'this makes no sense as an ini file'
        )
        try:
            config_manager.ConfigurationManager(
                (n,),
                argv_source=['--admin.conf=x.ini']
            )
            assert False, "where's the missing exception?"
        except AllHandlersFailedException, x:
            self.assertTrue('ConfigObj cannot load' in str(x))
        finally:
            os.remove('x.ini')

    #--------------------------------------------------------------------------
    def test_get_option_definitions(self):
        n = self._common_app_namespace_setup()
        n.add_option('silly', default=1)
        n.add_aggregation('strange', lambda x, y: 4)
        n.add_aggregation('weird', lambda x, y: 11)
        n.add_option('concrete', default=68)

        cm = config_manager.ConfigurationManager(
            (n,),
            argv_source=[]
        )
        opts = cm.get_option_names()
        for an_opt in opts:
            self.assertTrue(
                isinstance(cm.option_definitions[an_opt], Option)
            )
        self.assertEqual(len(opts), 9)  # there must be exactly 9 options

    #--------------------------------------------------------------------------
    @mock.patch('configman.config_manager.warnings')
    def test_warn_on_one_excess_options(self, mocked_warnings):
        assert 'configobj' in sys.modules.keys()

        n = self._common_app_namespace_setup()
        n.add_option('foo')
        open('x.ini', 'w').write(
            'foo=FOO\n'
            'bar=BAR\n'
        )
        try:
            config_manager.ConfigurationManager(
                (n,),
                argv_source=['--admin.conf=x.ini']
            )
            mocked_warnings.warn.assert_called_once_with(
                'Invalid options: bar'
            )
        finally:
            os.remove('x.ini')

    #--------------------------------------------------------------------------
    @mock.patch('configman.config_manager.warnings')
    def test_warn_on_multiple_excess_options(self, mocked_warnings):
        assert 'configobj' in sys.modules.keys()

        n = self._common_app_namespace_setup()
        n.add_option('foo')
        open('x.ini', 'w').write(
            'foo=FOO\n'
            'bar=BAR\n'
            'baz=BAZ\n'
        )
        try:
            config_manager.ConfigurationManager(
                (n,),
                argv_source=['--admin.conf=x.ini']
            )
            mocked_warnings.warn.assert_called_once_with(
                'Invalid options: bar, baz'
            )
        finally:
            os.remove('x.ini')

    #--------------------------------------------------------------------------
    def test_overlay_bug(self):
        # for Options that already exist and have been seen by the overlay
        # process, make sure that expanding a class doesn't just overwrite
        # the values back to their original defaults

        r = Namespace()
        r.add_option('fred', default=0)
        r.add_option(
            'class',
            default=int,
            from_string_converter=class_converter
        )

        # this class will bring in an Option that already exists called "fred".
        # Since overlay is done before expansion, "fred" should get set to 99.
        # Then expansion will bring in A's "fred" with a value of 77.  We need
        # make sure that the overlay process then puts its back to 99.
        class A(RequiredConfig):
            required_config = Namespace()
            required_config.add_option('fred', default='77')

        cm = config_manager.ConfigurationManager(
            [r],
            [{'fred': 21}, {'class': A}, {'fred': 99}]
        )
        cn = cm.get_config()
        self.assertEqual(cn.fred, 99)

    #--------------------------------------------------------------------------
    def test_overlay_reference_value_from_bug(self):
        # for Options that already exist and have been seen by the overlay
        # process, make sure that expanding a class doesn't just overwrite
        # the values back to their original defaults

        r = Namespace()
        r.add_option('fred', default=0, reference_value_from='a')
        r.add_option(
            'class',
            default=int,
            from_string_converter=class_converter
        )

        # this class will bring in an Option that already exists called "fred".
        # Since overlay is done before expansion, "fred" should get set to 99.
        # Then expansion will bring in A's "fred" with a value of 77.  We need
        # make sure that the overlay process then puts its back to 99.
        class A(RequiredConfig):
            required_config = Namespace()
            required_config.add_option(
                'fred',
                default='77',
                reference_value_from='a'
            )

        cm = config_manager.ConfigurationManager(
            [r],
            [{'a.fred': 21}, {'class': A}, {'a.fred': 99}]
        )
        cn = cm.get_config()
        self.assertEqual(cn.fred, 99)



# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is configman
#
# The Initial Developer of the Original Code is
# Mozilla Foundation
# Portions created by the Initial Developer are Copyright (C) 2011
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#    K Lars Lohn, lars@mozilla.com
#    Peter Bengtsson, peterbe@mozilla.com
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****

import dotdict
from option import Option, Aggregation


class Namespace(dotdict.DotDict):

    def __init__(self, doc=''):
        super(Namespace, self).__init__()
        object.__setattr__(self, '_doc', doc)  # force into attributes
        object.__setattr__(self, '_reference_value_from', False)

    #--------------------------------------------------------------------------
    def __setattr__(self, name, value):
        if isinstance(value, (Option, Namespace, Aggregation)):
            # then they know what they're doing already
            o = value
        else:
            o = Option(name=name, default=value, value=value)
        super(Namespace, self).__setattr__(name, o)

    #--------------------------------------------------------------------------
    def add_option(self, name, *args, **kwargs):
        """add an option to the namespace.   This can take two forms:
              'name' is a string representing the name of an option and the
              kwargs are its parameters, or 'name' is an instance of an Option
              object
        """
        if isinstance(name, Option):
            an_option = name
            name = an_option.name
        else:
            an_option = Option(name, *args, **kwargs)

        current_namespace = self
        name_parts = name.split('.')
        for a_path_component in name_parts[:-1]:
            if a_path_component not in current_namespace:
                current_namespace[a_path_component] = Namespace()
            current_namespace = current_namespace[a_path_component]
        an_option.name = name_parts[-1]

        setattr(current_namespace, an_option.name, an_option)

    #--------------------------------------------------------------------------
    def add_aggregation(self, name, function):
        an_aggregation = Aggregation(name, function)
        setattr(self, name, an_aggregation)

    #--------------------------------------------------------------------------
    def namespace(self, name, doc=''):
        setattr(self, name, Namespace(doc=doc))

    #--------------------------------------------------------------------------
    def set_value(self, name, value, strict=True):
        name_parts = name.split('.', 1)
        prefix = name_parts[0]
        try:
            candidate = getattr(self, prefix)
        except KeyError:
            if strict:
                raise
            candidate = Option(name)
            setattr(self, prefix, candidate)
        candidate_type = type(candidate)
        if candidate_type == Namespace:
            candidate.set_value(name_parts[1], value, strict)
        else:
            candidate.set_value(value)

    #--------------------------------------------------------------------------
    def safe_copy(self, reference_value_from=None):
        new_namespace = Namespace()
        if self._reference_value_from:
            new_namespace.ref_value_namespace()
        for key, opt in self.iteritems():
            if isinstance(opt, Option):
                new_namespace[key] = opt.copy()
                # assign a new reference_value if one has not been defined
                if not new_namespace[key].reference_value_from:
                    new_namespace[key].reference_value_from = \
                        reference_value_from
            elif isinstance(opt, Aggregation):
                new_namespace.add_aggregation(
                    opt.name,
                    opt.function
                )
            elif isinstance(opt, Namespace):
                new_namespace[key] = opt.safe_copy()
        return new_namespace

    #--------------------------------------------------------------------------
    def ref_value_namespace(self):
        """tag this namespace as having been created by the referenced value
        from system for collecting common resources together."""
        # awkward syntax - because the base class DotDict hijacks the
        # the __setattr__ method, this is the only way to actually force a
        # value to become an attribute rather than member of the dict
        object.__setattr__(self, '_reference_value_from', True)




# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is configman
#
# The Initial Developer of the Original Code is
# Mozilla Foundation
# Portions created by the Initial Developer are Copyright (C) 2011
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#    K Lars Lohn, lars@mozilla.com
#    Peter Bengtsson, peterbe@mozilla.com
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****

import sys
import os
import collections
import inspect
import os.path
import contextlib
import functools
import warnings

import configman as cm
import converters as conv
import config_exceptions as exc
import value_sources
import def_sources

#==============================================================================
# for convenience define some external symbols here
from option import Option, Aggregation
from dotdict import DotDict, DotDictWithAcquisition
from namespace import Namespace
from required_config import RequiredConfig
from config_file_future_proxy import ConfigFileFutureProxy


#==============================================================================
class ConfigurationManager(object):

    #--------------------------------------------------------------------------
    def __init__(
        self,
        definition_source=None,
        values_source_list=None,
        argv_source=None,
        #use_config_files=True,
        use_auto_help=True,
        use_admin_controls=True,
        quit_after_admin=True,
        options_banned_from_help=None,
        app_name='',
        app_version='',
        app_description='',
        config_pathname='.',
        config_optional=True,
    ):
        """create and initialize a configman object.

        parameters:
          definition_source - a namespace or list of namespaces from which
                              configman is to fetch the definitions of the
                              configuration parameters.
          values_source_list - (optional) a hierarchical list of sources for
                               values for the configuration parameters.
                               As values are copied from these sources,
                               conficting values are resolved with sources
                               on the right getting preference over sources on
                               the left.
          argv_source - if the values_source_list contains a commandline
                        source, this value is an alternative source for
                        actual command line arguments.  Useful for testing or
                        preprocessing command line arguments.
          use_auto_help - set to True if configman is to automatically set up
                          help output for command line invocations.
          use_admin_controls - configman can add command line flags that it
                               interprets independently of the app defined
                               arguments.  True enables this capability, while,
                               False supresses it.
          quit_after_admin - if True and admin controls are enabled and used,
                             call sys.exit to end the app.  This is useful to
                             stop the app from running if all that was done
                             was to write a config file or stop after help.
          options_banned_from_help - a list of strings that will censor the
                                     output of help to prevent specified
                                    options from being listed in the help
                                    output.  This is useful for hiding debug
                                    or secret command line arguments.
          app_name - assigns a name to the app.  This is used in help output
                     and as a default basename for config files.
          app_version - assigns a version for the app used help output.
          app_description - assigns a description for the app to be used in
                            the help output.
          config_pathname - a hard coded path to the directory of or the full
                            path and name of the configuration file.
          config_optional - a boolean indicating if a missing default config
                            file is optional.  Note: this is only for the
                            default config file.  If a config file is specified
                            on the commandline, it _must_ exsist."""

        # instead of allowing mutables as default keyword argument values...
        if definition_source is None:
            definition_source_list = []
        elif (
            isinstance(definition_source, collections.Sequence) and
            not isinstance(definition_source, basestring)
        ):
            definition_source_list = list(definition_source)
        else:
            definition_source_list = [definition_source]

        if argv_source is None:
            self.argv_source = sys.argv[1:]
            self.app_invocation_name = sys.argv[0]
        else:
            self.argv_source = argv_source
            self.app_invocation_name = app_name
        if options_banned_from_help is None:
            options_banned_from_help = ['application']
        self.config_pathname = config_pathname
        self.config_optional = config_optional

        self.app_name = app_name
        self.app_version = app_version
        self.app_description = app_description

        self.args = []  # extra commandline arguments that are not switches
                        # will be stored here.

        self._config = None  # eventual container for DOM-like config object

        self.option_definitions = Namespace()
        self.definition_source_list = definition_source_list

        if values_source_list is None:
            # nothing set, assume defaults
            if use_admin_controls:
                values_source_list = (
                    cm.ConfigFileFutureProxy,
                    cm.environment,
                    cm.command_line
                )
            else:
                values_source_list = (
                    cm.environment,
                    cm.command_line
                )

        admin_tasks_done = False
        self.admin_controls_list = [
            'help',
            'admin.conf',
            'admin.dump_conf',
            'admin.print_conf',
            'admin.strict'
        ]
        self.options_banned_from_help = options_banned_from_help

        if use_auto_help:
            self._setup_auto_help()
        if use_admin_controls:
            admin_options = self._setup_admin_options(values_source_list)
            self.definition_source_list.append(admin_options)

        # iterate through the option definitions to create the nested dict
        # hierarchy of all the options called 'option_definitions'
        for a_definition_source in self.definition_source_list:
            try:
                safe_copy_of_def_source = a_definition_source.safe_copy()
            except AttributeError:
                # apparently, the definition source was not in the form of a
                # Namespace object.  This isn't a show stopper, but we don't
                # know how to make a copy of this object safely: we know from
                # experience that the stock copy.copy method leads to grief
                # as many sub-objects within an option definition source can
                # not be copied that way (classes, for example).
                # The only action we can take is to trust and continue with the
                # original copy of the definition source.
                safe_copy_of_def_source = a_definition_source
            def_sources.setup_definitions(
                safe_copy_of_def_source,
                self.option_definitions
            )

        if use_admin_controls:
            # the name of the config file needs to be loaded from the command
            # line prior to processing the rest of the command line options.
            config_filename = \
                value_sources.config_filename_from_commandline(self)
            if (
                config_filename
                and cm.ConfigFileFutureProxy in values_source_list
            ):
                self.option_definitions.admin.conf.default = config_filename

        self.values_source_list = value_sources.wrap(
            values_source_list,
            self
        )

        known_keys = self._overlay_expand()
        self._check_for_mismatches(known_keys)

        # the app_name, app_version and app_description are to come from
        # if 'application' option if it is present. If it is not present,
        # get the app_name,et al, from parameters passed into the constructor.
        # if those are empty, set app_name, et al, to empty strings
        try:
            app_option = self._get_option('application')
            self.app_name = getattr(app_option.value, 'app_name', '')
            self.app_version = getattr(app_option.value, 'app_version', '')
            self.app_description = getattr(
                app_option.value,
                'app_description',
                ''
            )
        except exc.NotAnOptionError:
            # there is no 'application' option, continue to use the
            # 'app_name' from the parameters passed in, if they exist.
            pass

        if use_auto_help and self._get_option('help').value:
            self.output_summary()
            admin_tasks_done = True

        if use_admin_controls and self._get_option('admin.print_conf').value:
            self.print_conf()
            admin_tasks_done = True

        if use_admin_controls and self._get_option('admin.dump_conf').value:
            self.dump_conf()
            admin_tasks_done = True

        if quit_after_admin and admin_tasks_done:
            sys.exit()

    #--------------------------------------------------------------------------
    @contextlib.contextmanager
    def context(self):
        """return a config as a context that calls close on every item when
        it goes out of scope"""
        config = None
        try:
            config = self.get_config()
            yield config
        finally:
            if config:
                self._walk_and_close(config)

    #--------------------------------------------------------------------------
    def get_config(self, mapping_class=DotDictWithAcquisition):
        config = self._generate_config(mapping_class)
        if self._aggregate(self.option_definitions, config, config):
            # state changed, must regenerate
            return self._generate_config(mapping_class)
        else:
            return config

    #--------------------------------------------------------------------------
    def output_summary(self,
                       output_stream=sys.stdout,
                       block_password=True):
        """outputs a usage tip and the list of acceptable commands.
        This is useful as the output of the 'help' option.

        parameters:
            output_stream - an open file-like object suitable for use as the
                            target of a print statement
            block_password - a boolean driving the use of a string of * in
                             place of the value for any object containing the
                             substring 'passowrd'
        """
        if self.app_name or self.app_description:
            print >> output_stream, 'Application:',
        if self.app_name:
            print >> output_stream, self.app_name, self.app_version
        if self.app_description:
            print >> output_stream, self.app_description
        if self.app_name or self.app_description:
            print >> output_stream, ''

        names_list = self.get_option_names()
        print >> output_stream, (
            "usage:\n",
            self.app_invocation_name,
            "[OPTIONS]..."
        ),
        bracket_count = 0
        for key in names_list:
            an_option = self.option_definitions[key]
            if an_option.is_argument:
                if an_option.default is None:
                    print >> output_stream, an_option.name,
                else:
                    print >> output_stream, "[ %s" % an_option.name,
                    bracket_count += 1
        print >> output_stream, ']' * bracket_count, '\n'

        names_list.sort()
        if names_list:
            print >> output_stream, 'OPTIONS:'

        pad = ' ' * 4

        for name in names_list:
            if name in self.options_banned_from_help:
                continue
            option = self._get_option(name)

            line = ' ' * 2  # always start with 2 spaces
            if option.short_form:
                line += '-%s, ' % option.short_form
            line += '--%s' % name
            line += '\n'

            doc = option.doc if option.doc is not None else ''
            if doc:
                line += '%s%s\n' % (pad, doc)
            try:
                value = option.value
                type_of_value = type(value)
                converter_function = conv.to_string_converters[type_of_value]
                default = converter_function(value)
            except KeyError:
                default = option.value
            if default is not None:
                if 'password' in name.lower():
                    default = '*********'
                if name not in ('help',):
                    # don't bother with certain dead obvious ones
                    line += '%s(default: %s)\n' % (pad, default)

            print >> output_stream, line

    #--------------------------------------------------------------------------
    def print_conf(self):
        """write a config file to the pathname specified in the parameter.  The
        file extention determines the type of file written and must match a
        registered type.

        parameters:
            config_pathname - the full path and filename of the target config
                               file."""

        config_file_type = self._get_option('admin.print_conf').value

        @contextlib.contextmanager
        def stdout_opener():
            yield sys.stdout

        skip_keys = [
            k for (k, v)
            in self.option_definitions.iteritems()
            if isinstance(v, Option) and v.exclude_from_print_conf
        ]
        self.write_conf(config_file_type, stdout_opener, skip_keys=skip_keys)

    #--------------------------------------------------------------------------
    def dump_conf(self, config_pathname=None):
        """write a config file to the pathname specified in the parameter.  The
        file extention determines the type of file written and must match a
        registered type.

        parameters:
            config_pathname - the full path and filename of the target config
                               file."""

        if not config_pathname:
            config_pathname = self._get_option('admin.dump_conf').value

        opener = functools.partial(open, config_pathname, 'w')
        config_file_type = os.path.splitext(config_pathname)[1][1:]

        skip_keys = [
            k for (k, v)
            in self.option_definitions.iteritems()
            if isinstance(v, Option) and v.exclude_from_dump_conf
        ]

        self.write_conf(config_file_type, opener, skip_keys=skip_keys)

    #--------------------------------------------------------------------------
    def write_conf(self, config_file_type, opener, skip_keys=None):
        """write a configuration file to a file-like object.

        parameters:
            config_file_type - a string containing a registered file type OR
                               a for_XXX module from the value_source
                               package.  Passing in an string that is
                               unregistered will result in a KeyError
            opener - a callable object or function that returns a file like
                     object that works as a context in a with statement."""

        blocked_keys = self.admin_controls_list
        if skip_keys:
            blocked_keys.extend(skip_keys)

        if blocked_keys:
            option_defs = self.option_definitions.safe_copy()
            for a_blocked_key in blocked_keys:
                try:
                    del option_defs[a_blocked_key]
                except (AttributeError, KeyError):
                    # okay that key isn't here
                    pass
            # remove empty namespaces
            all_keys = [k for k in
                        option_defs.keys_breadth_first(include_dicts=True)]
            for key in all_keys:
                candidate = option_defs[key]
                if (isinstance(candidate, Namespace) and not len(candidate)):
                    del option_defs[key]
        else:
            option_defs = self.option_definitions

        value_sources.write(config_file_type,
                            option_defs,
                            opener)

    #--------------------------------------------------------------------------
    def log_config(self, logger):
        """write out the current configuration to a log-like object.

        parameters:
            logger - a object that implements a method called 'info' with the
                     same semantics as the call to 'logger.info'"""

        logger.info("app_name: %s", self.app_name)
        logger.info("app_version: %s", self.app_version)
        logger.info("current configuration:")
        config = [(key, self.option_definitions[key].value)
                  for key in self.option_definitions.keys_breadth_first()
                  if key not in self.admin_controls_list]
        config.sort()
        for key, val in config:
            if 'password' in key.lower():
                logger.info('%s: *********', key)
            else:
                try:
                    logger.info('%s: %s', key,
                                conv.to_string_converters[type(key)](val))
                except KeyError:
                    logger.info('%s: %s', key, val)

    #--------------------------------------------------------------------------
    def get_option_names(self):
        """returns a list of fully qualified option names.

        returns:
            a list of strings representing the Options in the source Namespace
            list.  Each item will be fully qualified with dot delimited
            Namespace names.
        """
        return [x for x in self.option_definitions.keys_breadth_first()
                if isinstance(self.option_definitions[x], Option)]

    #--------------------------------------------------------------------------
    def _create_reference_value_from_links(self, keys, known_keys):
        """this method steps through the option definitions looking for
        alt paths.  On finding one, it creates the 'reference_value_from' links
        within the option definitions and populates it with copied options."""
        # a set of known reference_value_from_links
        set_of_reference_value_from_links = set()
        for key in (k for k in keys if k not in known_keys):

            an_option = self.option_definitions[key]
            if an_option.reference_value_from:

                reference_name = '.'.join((
                    an_option.reference_value_from,
                    an_option.name
                ))
                if reference_name in self.option_definitions:
                    continue  # this referenced value has already been defined
                              # no need to repeat it - skip on to the next key
                reference_option = an_option.copy()
                reference_option.reference_value_from = None
                reference_option.name = reference_name
                # wait, aren't we setting a fully qualified dotted name into
                # the name field?  Yes, 'add_option' below sees that
                # full pathname and does the right thing with it to ensure
                # that the reference_option is created within the
                # correct namespace
                set_of_reference_value_from_links.add(reference_option.name)
                self.option_definitions.add_option(reference_option)

        for a_reference_value_from in set_of_reference_value_from_links:
            for x in range(a_reference_value_from.count('.')):
                namespace_path = a_reference_value_from.rsplit('.', x + 1)[0]
                self.option_definitions[namespace_path].ref_value_namespace()

        return set_of_reference_value_from_links

    #--------------------------------------------------------------------------
    def _overlay_expand(self):
        """This method overlays each of the value sources onto the default
        in each of the defined options.  It does so using a breadth first
        iteration, overlaying and expanding each level of the tree in turn.
        As soon as no changes were made to any level, the loop breaks and the
        work is done.  The actual action of the overlay is to take the value
        from the source and copy into the 'default' member of each Option
        object.

        "expansion" means converting an option value into its real type from
        string. The conversion is accomplished by simply calling the
        'set_value' method of the Option object.  If the resultant type has its
        own configuration options, bring those into the current namespace and
        then proceed to overlay/expand those.
        """
        new_keys_discovered = True  # loop control, False breaks the loop
        known_keys = set()  # a set of keys that have been expanded

        while new_keys_discovered:  # loop until nothing more is done
            # keys holds a list of all keys in the option definitons in
            # breadth first order using this form: [ 'x', 'y', 'z', 'x.a',
            # 'x.b', 'z.a', 'z.b', 'x.a.j', 'x.a.k', 'x.b.h']
            keys = [
                x for x
                in self.option_definitions.keys_breadth_first()
                if isinstance(self.option_definitions[x], Option)
            ]
            new_keys_discovered = False  # setup to break loop

            # create alternate paths options
            set_of_reference_value_from_links = \
                self._create_reference_value_from_links(
                    keys,
                    known_keys
                )
            all_keys = list(set_of_reference_value_from_links) + keys

            # overlay process:
            # fetch all the default values from the value sources before
            # applying the from string conversions
            #

            for key in (k for k in all_keys if k not in known_keys):
                #if not isinstance(an_option, Option):
                #   continue  # aggregations and other types are ignored
                # loop through all the value sources looking for values
                # that match this current key.
                if self.option_definitions[key].reference_value_from:
                    reference_value_from = (
                        self.option_definitions[key].reference_value_from
                    )
                    top_key = key.split('.')[-1]
                    self.option_definitions[key].default = (
                        self.option_definitions[reference_value_from]
                        [top_key].default
                    )
                for a_value_source in self.values_source_list:
                    try:
                        # get all the option values from this value source
                        val_src_dict = a_value_source.get_values(
                            self,
                            True
                        )
                        # make sure it is in the form of a DotDict
                        if not isinstance(val_src_dict, DotDict):
                            val_src_dict = (
                                DotDictWithAcquisition(val_src_dict)
                            )
                        # get the Option for this key
                        opt = self.option_definitions[key]
                        # overlay the default with the new value from
                        # the value source.  This assignment may come
                        # via acquisition, so the key given may not have
                        # been an exact match for what was returned.
                        opt.default = val_src_dict[key]
                    except KeyError, x:
                        pass  # okay, that source doesn't have this value

            # expansion process:
            # step through all the keys converting them to their proper
            # types and bringing in any new keys in the process
            for key in (k for k in all_keys if k not in known_keys):
                # mark this key as having been seen and processed
                known_keys.add(key)
                an_option = self.option_definitions[key]
                #if not isinstance(an_option, Option):
                #    continue  # aggregations, namespaces are ignored
                # apply the from string conversion to make the real value
                an_option.set_value(an_option.default)
                # new values have been seen, don't let loop break
                new_keys_discovered = True
                try:
                    try:
                        # try to fetch new requirements from this value
                        new_req = an_option.value.get_required_config()
                    except AttributeError:
                        new_req = an_option.value.required_config
                    # make sure what we got as new_req is actually a
                    # Mapping of some sort
                    if not isinstance(new_req, collections.Mapping):
                        # we didn't get a mapping, perhaps the option value
                        # was a Mock object - in any case we can't try to
                        # interpret 'new_req' as a configman requirement
                        # collection.  We must abandon processing this
                        # option further
                        continue
                    # get the parent namespace
                    current_namespace = self.option_definitions.parent(key)
                    if current_namespace is None:
                        # we're at the top level, use the base namespace
                        current_namespace = self.option_definitions
                    # some new Options to be brought in may have already been
                    # seen and in the known_keys set.  They must be marked
                    # as unseen so that the new default doesn't overwrite any
                    # of the overlays that have already taken place.
                    known_keys = known_keys.difference(
                        known_keys.intersection(new_req.keys())
                    )
                    # add the new Options to the namespace
                    current_namespace.update(new_req.safe_copy(
                        an_option.reference_value_from
                    ))
                except AttributeError, x:
                    # there are apparently no new Options to bring in from
                    # this option's value
                    pass
        return known_keys

    #--------------------------------------------------------------------------
    def _check_for_mismatches(self, known_keys):
        """check for bad options from value sources"""
        for a_value_source in self.values_source_list:
            try:
                if a_value_source.always_ignore_mismatches:
                    continue
            except AttributeError:
                # ok, this values source doesn't have the concept
                # always igoring mismatches, we won't tolerate mismatches
                pass
            # we want to fetch the keys from the value sources so that we can
            # check for mismatches.  Commandline value sources, are different,
            # we never want to allow unmatched keys from the command line.
            # By detecting if this value source is a command line source, we
            # can employ the command line's own mismatch detection.  The
            # boolean 'allow_mismatches' controls application of the tollerance
            # for mismatches.
            if hasattr(a_value_source, 'command_line_value_source'):
                allow_mismatches = False
            else:
                allow_mismatches = True
            # make a set of all the keys from a value source in the form
            # of strings like this: 'x.y.z'
            value_source_mapping = a_value_source.get_values(
                self,
                allow_mismatches
            )
            value_source_keys_set = set([
                k for k in
                DotDict(value_source_mapping).keys_breadth_first()
            ])
            # make a set of the keys that didn't match any of the known
            # keys in the requirements
            unmatched_keys = value_source_keys_set.difference(known_keys)
            # some of the unmatched keys may actually be ok because the were
            # used during acquisition.
            # remove keys of the form 'y.z' if they match a known key of the
            # form 'x.y.z'
            for key in unmatched_keys.copy():
                key_is_okay = reduce(
                    lambda x, y: x or y,
                    (known_key.endswith(key) for known_key in known_keys)
                )
                if key_is_okay:
                    unmatched_keys.remove(key)
            # anything left in the unmatched_key set is a badly formed key.
            # issue a warning
            if unmatched_keys:
                if self.option_definitions.admin.strict.default:
                    # raise hell...
                    if len(unmatched_keys) > 1:
                        raise exc.NotAnOptionError(
                            "%s are not valid Options" % unmatched_keys
                        )
                    elif len(unmatched_keys) == 1:
                        raise exc.NotAnOptionError(
                            "%s is not a valid Option" % unmatched_keys.pop()
                        )
                else:
                    warnings.warn(
                        'Invalid options: %s' % ', '.join(unmatched_keys)
                    )

    #--------------------------------------------------------------------------
    @staticmethod
    def _walk_and_close(a_dict):
        for val in a_dict.itervalues():
            if isinstance(val, collections.Mapping):
                ConfigurationManager._walk_and_close(val)
            if hasattr(val, 'close') and not inspect.isclass(val):
                val.close()

    #--------------------------------------------------------------------------
    def _generate_config(self, mapping_class):
        """This routine generates a copy of the DotDict based config"""
        config = mapping_class()
        self._walk_config_copy_values(
            self.option_definitions,
            config,
            mapping_class
        )
        return config

    #--------------------------------------------------------------------------
    def _setup_auto_help(self):
        help_option = Option(name='help', doc='print this', default=False)
        self.definition_source_list.append({'help': help_option})

    #--------------------------------------------------------------------------
    def _get_config_pathname(self):
        if os.path.isdir(self.config_pathname):
            # we've got a path with no file name at the end
            # use the appname as the file name and default to an 'ini'
            # config file type
            if self.app_name:
                return os.path.join(
                    self.config_pathname,
                    '%s.ini' % self.app_name
                )
            else:
                # there is no app_name yet
                # we'll decline to return anything
                return None
        return self.config_pathname

    #--------------------------------------------------------------------------
    def _setup_admin_options(self, values_source_list):
        base_namespace = Namespace()
        base_namespace.admin = admin = Namespace()
        admin.add_option(
            name='print_conf',
            default=None,
            doc='write current config to stdout (%s)'
                % ', '.join(value_sources.file_extension_dispatch.keys())
        )
        admin.add_option(
            name='dump_conf',
            default='',
            doc='a pathname to which to write the current config',
        )
        admin.add_option(
            name='strict',
            default=False,
            doc='mismatched options generate exceptions rather'
                ' than just warnings'
        )
        # only offer the config file admin options if they've been requested in
        # the values source list
        if ConfigFileFutureProxy in values_source_list:
            default_config_pathname = self._get_config_pathname()
            admin.add_option(
                name='conf',
                default=default_config_pathname,
                doc='the pathname of the config file (path/filename)',
            )
        return base_namespace

    #--------------------------------------------------------------------------
    def _walk_config_copy_values(self, source, destination, mapping_class):
        for key, val in source.items():
            value_type = type(val)
            if isinstance(val, Option) or isinstance(val, Aggregation):
                destination[key] = val.value
            elif value_type == Namespace:
                destination[key] = d = mapping_class()
                self._walk_config_copy_values(val, d, mapping_class)

    #--------------------------------------------------------------------------
    def _aggregate(self, source, base_namespace, local_namespace):
        aggregates_found = False
        for key, val in source.items():
            if isinstance(val, Namespace):
                new_aggregates_found = self._aggregate(
                    val,
                    base_namespace,
                    local_namespace[key]
                )
                aggregates_found = new_aggregates_found or aggregates_found
            elif isinstance(val, Aggregation):
                val.aggregate(base_namespace, local_namespace, self.args)
                aggregates_found = True
            # skip Options, we're only dealing with Aggregations
        return aggregates_found

    #--------------------------------------------------------------------------
    @staticmethod
    def _option_sort(x_tuple):
        key, val = x_tuple
        if isinstance(val, Namespace):
            return 'zzzzzzzzzzz%s' % key
        else:
            return key

    #--------------------------------------------------------------------------
    @staticmethod
    def _block_password(qkey, key, value, block_password=True):
        if block_password and 'password' in key.lower():
            value = '*********'
        return qkey, key, value

    #--------------------------------------------------------------------------
    def _get_option(self, name):
        try:
            return self.option_definitions[name]
        except KeyError:
            raise exc.NotAnOptionError('%s is not a known option name' % name)

    #--------------------------------------------------------------------------
    def _get_options(self, source=None, options=None, prefix=''):
        return [
            (key, self.option_definitions[key])
            for key in self.option_definitions.keys_breadth_first()
        ]



# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is configman
#
# The Initial Developer of the Original Code is
# Mozilla Foundation
# Portions created by the Initial Developer are Copyright (C) 2011
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#    K Lars Lohn, lars@mozilla.com
#    Peter Bengtsson, peterbe@mozilla.com
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****

import json
import collections
import sys

from .. import converters as conv
from ..namespace import Namespace
from ..option import Option, Aggregation

from source_exceptions import (ValueException, NotEnoughInformationException,
                               CantHandleTypeException)

can_handle = (
    basestring,
    json
)

file_name_extension = 'json'


#==============================================================================
class LoadingJsonFileFailsException(ValueException):
    pass


#==============================================================================
class ValueSource(object):

    #--------------------------------------------------------------------------
    def __init__(self, source, the_config_manager=None):
        self.values = None
        if source is json:
            try:
                app = the_config_manager._get_option('admin.application')
                source = "%s.%s" % (app.value.app_name, file_name_extension)
            except (AttributeError, KeyError):
                raise NotEnoughInformationException(
                    "Can't setup an json file without knowing the file name"
                )
        if (
            isinstance(source, basestring)
            and source.endswith(file_name_extension)
        ):
            try:
                with open(source) as fp:
                    self.values = json.load(fp)
            except IOError, x:
                # The file doesn't exist.  That's ok, we'll give warning
                # but this isn't a fatal error
                import warnings
                warnings.warn("%s doesn't exist" % source)
                self.values = {}
            except ValueError:
                raise LoadingJsonFileFailsException(
                    "Cannot load json: %s" % str(x)
                )
        else:
            raise CantHandleTypeException()

    #--------------------------------------------------------------------------
    def get_values(self, config_manager, ignore_mismatches):
        return self.values

    #--------------------------------------------------------------------------
    @staticmethod
    def recursive_default_dict():
        return collections.defaultdict(ValueSource.recursive_default_dict)

    #--------------------------------------------------------------------------
    @staticmethod
    def write(source_dict, output_stream=sys.stdout):
        json_dict = ValueSource.recursive_default_dict()
        for qkey in source_dict.keys_breadth_first(include_dicts=True):
            val = source_dict[qkey]
            if isinstance(val, Namespace):
                continue
            d = json_dict
            for x in qkey.split('.'):
                d = d[x]
            if isinstance(val, Option):
                for okey, oval in val.__dict__.iteritems():
                    try:
                        d[okey] = conv.to_string_converters[type(oval)](oval)
                    except KeyError:
                        d[okey] = str(oval)
                d['default'] = d['value']
            elif isinstance(val, Aggregation):
                d['name'] = val.name
                fn = val.function
                d['function'] = conv.to_string_converters[type(fn)](fn)
        json.dump(json_dict, output_stream)


